@@ -6,7 +6,6 @@ try:
 except DistributionNotFound:
     pass
 
-
 from . import differentiation
 from . import feature_library
 from . import optimizers
@@ -18,16 +17,23 @@ from .differentiation import FiniteDifference
 from .differentiation import SINDyDerivative
 from .differentiation import SmoothedFiniteDifference
 from .feature_library import ConcatLibrary
+from .feature_library import TensoredLibrary
+from .feature_library import GeneralizedLibrary
 from .feature_library import CustomLibrary
 from .feature_library import FourierLibrary
 from .feature_library import IdentityLibrary
 from .feature_library import PolynomialLibrary
+from .feature_library import PDELibrary
+from .feature_library import SINDyPILibrary
 from .optimizers import BaseOptimizer
 from .optimizers import ConstrainedSR3
-from .optimizers import TrappingSR3
+from .optimizers import FROLS
 from .optimizers import SINDyOptimizer
 from .optimizers import SR3
+from .optimizers import SSR
 from .optimizers import STLSQ
+from .optimizers import SINDyPI
+from .optimizers import TrappingSR3
 
 
 __all__ = ["SINDy"]
