@@ -2,6 +2,9 @@ from itertools import combinations
 from itertools import combinations_with_replacement as combinations_w_r
 
 from numpy import empty
+from numpy import ones
+from numpy import shape
+from sklearn import __version__
 from sklearn.utils import check_array
 from sklearn.utils.validation import check_is_fitted
 
@@ -14,8 +17,10 @@ class CustomLibrary(BaseFeatureLibrary):
     Parameters
     ----------
     library_functions : list of mathematical functions
-        Functions to include in the library. Each function will be
-        applied to each input variable.
+        Functions to include in the library. Default is to use same functions
+        for all variables. Can also be used so that each variable has an
+        associated library, in this case library_functions is shape
+        (n_input_features, num_library_functions)
 
     function_names : list of functions, optional (default None)
         List of functions used to generate feature names for each library
@@ -23,9 +28,10 @@ class CustomLibrary(BaseFeatureLibrary):
         a variable name), and output a string depiction of the respective
         mathematical function applied to that variable. For example, if the
         first library function is sine, the name function might return
-        :math:`\\sin(x)` given :math:`x` as input. The function_names list must be the
-        same length as library_functions. If no list of function names is
-        provided, defaults to using :math:`[ f_0(x),f_1(x), f_2(x), \\ldots ]`.
+        :math:`\\sin(x)` given :math:`x` as input. The function_names list
+        must be the same length as library_functions. If no list of function
+        names is provided, defaults to using
+        :math:`[ f_0(x),f_1(x), f_2(x), \\ldots ]`.
 
     interaction_only : boolean, optional (default True)
         Whether to omit self-interaction terms.
@@ -34,6 +40,20 @@ class CustomLibrary(BaseFeatureLibrary):
         will be included.
         If False, all combinations will be included.
 
+    library_ensemble : boolean, optional (default False)
+        Whether or not to use library bagging (regress on subset of the
+        candidate terms in the library)
+
+    ensemble_indices : integer array, optional (default [0])
+        The indices to use for ensembling the library.
+
+    include_bias : boolean, optional (default False)
+        If True (default), then include a bias column, the feature in which
+        all polynomial powers are zero (i.e. a column of ones - acts as an
+        intercept term in a linear model).
+        This is hard to do with just lambda functions, because if the system
+        is not 1D, lambdas will generate duplicates.
+
     Attributes
     ----------
     functions : list of functions
@@ -45,6 +65,8 @@ class CustomLibrary(BaseFeatureLibrary):
 
     n_input_features_ : int
         The total number of input features.
+        WARNING: This is deprecated in scikit-learn version 1.0 and higher so
+        we check the sklearn.__version__ and switch to n_features_in if needed.
 
     n_output_features_ : int
         The total number of output features. The number of output features
@@ -66,15 +88,28 @@ class CustomLibrary(BaseFeatureLibrary):
     ['f0(x0)', 'f0(x1)', 'f1(x0,x1)']
     """
 
-    def __init__(self, library_functions, function_names=None, interaction_only=True):
-        super(CustomLibrary, self).__init__()
+    def __init__(
+        self,
+        library_functions,
+        function_names=None,
+        interaction_only=True,
+        library_ensemble=False,
+        ensemble_indices=[0],
+        include_bias=False,
+    ):
+        super(CustomLibrary, self).__init__(
+            library_ensemble=library_ensemble, ensemble_indices=ensemble_indices
+        )
         self.functions = library_functions
         self.function_names = function_names
-        if function_names and (len(library_functions) != len(function_names)):
+        if function_names and (
+            shape(library_functions)[-1] != shape(function_names)[-1]
+        ):
             raise ValueError(
                 "library_functions and function_names must have the same"
                 " number of elements"
             )
+        self.include_bias = include_bias
         self.interaction_only = interaction_only
 
     @staticmethod
@@ -97,12 +132,19 @@ class CustomLibrary(BaseFeatureLibrary):
         output_feature_names : list of string, length n_output_features
         """
         check_is_fitted(self)
+
+        if float(__version__[:3]) >= 1.0:
+            n_input_features = self.n_features_in_
+        else:
+            n_input_features = self.n_input_features_
         if input_features is None:
-            input_features = ["x%d" % i for i in range(self.n_input_features_)]
+            input_features = ["x%d" % i for i in range(n_input_features)]
         feature_names = []
+        if self.include_bias:
+            feature_names.append("1")
         for i, f in enumerate(self.functions):
             for c in self._combinations(
-                self.n_input_features_, f.__code__.co_argcount, self.interaction_only
+                n_input_features, f.__code__.co_argcount, self.interaction_only
             ):
                 feature_names.append(
                     self.function_names[i](*[input_features[j] for j in c])
@@ -122,13 +164,18 @@ class CustomLibrary(BaseFeatureLibrary):
         self : instance
         """
         n_samples, n_features = check_array(x).shape
-        self.n_input_features_ = n_features
+        if float(__version__[:3]) >= 1.0:
+            self.n_features_in_ = n_features
+        else:
+            self.n_input_features_ = n_features
         n_output_features = 0
         for f in self.functions:
             n_args = f.__code__.co_argcount
             n_output_features += len(
                 list(self._combinations(n_features, n_args, self.interaction_only))
             )
+        if self.include_bias:
+            n_output_features += 1
         self.n_output_features_ = n_output_features
         if self.function_names is None:
             self.function_names = list(
@@ -159,16 +206,25 @@ class CustomLibrary(BaseFeatureLibrary):
 
         n_samples, n_features = x.shape
 
-        if n_features != self.n_input_features_:
+        if float(__version__[:3]) >= 1.0:
+            n_input_features = self.n_features_in_
+        else:
+            n_input_features = self.n_input_features_
+
+        if n_features != n_input_features:
             raise ValueError("x shape does not match training shape")
 
         xp = empty((n_samples, self.n_output_features_), dtype=x.dtype)
         library_idx = 0
+        if self.include_bias:
+            xp[:, library_idx] = ones(n_samples)
+            library_idx += 1
         for f in self.functions:
             for c in self._combinations(
-                self.n_input_features_, f.__code__.co_argcount, self.interaction_only
+                n_input_features, f.__code__.co_argcount, self.interaction_only
             ):
                 xp[:, library_idx] = f(*[x[:, j] for j in c])
                 library_idx += 1
 
-        return xp
+        # If library bagging, return xp missing the terms at ensemble_indices
+        return self._ensemble(xp)
