@@ -1,7 +1,10 @@
 from .base import BaseOptimizer
 from .constrained_sr3 import ConstrainedSR3
+from .frols import FROLS
 from .sindy_optimizer import SINDyOptimizer
+from .sindy_pi import SINDyPI
 from .sr3 import SR3
+from .ssr import SSR
 from .stlsq import STLSQ
 from .trapping_sr3 import TrappingSR3
 
@@ -12,4 +15,7 @@ __all__ = [
     "STLSQ",
     "ConstrainedSR3",
     "TrappingSR3",
+    "SSR",
+    "FROLS",
+    "SINDyPI",
 ]
