@@ -1,15 +1,22 @@
 import warnings
 from typing import Sequence
 
+import numpy.random
 from numpy import concatenate
+from numpy import copy
+from numpy import insert
 from numpy import isscalar
+from numpy import median
 from numpy import ndim
 from numpy import newaxis
+from numpy import sort
 from numpy import vstack
 from numpy import zeros
 from scipy.integrate import odeint
+from scipy.integrate import solve_ivp
 from scipy.interpolate import interp1d
 from scipy.linalg import LinAlgWarning
+from sklearn import __version__
 from sklearn.base import BaseEstimator
 from sklearn.exceptions import ConvergenceWarning
 from sklearn.metrics import r2_score
@@ -17,10 +24,13 @@ from sklearn.pipeline import Pipeline
 from sklearn.utils.validation import check_is_fitted
 
 from .differentiation import FiniteDifference
+from .feature_library import PDELibrary
 from .feature_library import PolynomialLibrary
+from .feature_library import SINDyPILibrary
 from .optimizers import SINDyOptimizer
 from .optimizers import STLSQ
 from .utils import drop_nan_rows
+from .utils import drop_random_rows
 from .utils import equations
 from .utils import validate_control_variables
 from .utils import validate_input
@@ -80,13 +90,13 @@ class SINDy(BaseEstimator):
     Examples
     --------
     >>> import numpy as np
-    >>> from scipy.integrate import odeint
+    >>> from scipy.integrate import solve_ivp
     >>> from pysindy import SINDy
     >>> lorenz = lambda z,t : [10*(z[1] - z[0]),
     >>>                        z[0]*(28 - z[2]) - z[1],
     >>>                        z[0]*z[1] - 8/3*z[2]]
     >>> t = np.arange(0,2,.002)
-    >>> x = odeint(lorenz, [-8,8,27], t)
+    >>> x = solve_ivp(lorenz, [-8,8,27], t)
     >>> model = SINDy()
     >>> model.fit(x, t=t[1]-t[0])
     >>> model.print()
@@ -108,7 +118,7 @@ class SINDy(BaseEstimator):
     0.999999985520653
 
     >>> import numpy as np
-    >>> from scipy.integrate import odeint
+    >>> from scipy.integrate import solve_ivp
     >>> from pysindy import SINDy
     >>> u = lambda t : np.sin(2 * t)
     >>> lorenz_c = lambda z,t : [
@@ -117,7 +127,7 @@ class SINDy(BaseEstimator):
                 z[0] * z[1] - 8 / 3 * z[2],
         ]
     >>> t = np.arange(0,2,0.002)
-    >>> x = odeint(lorenz_c, [-8,8,27], t)
+    >>> x = solve_ivp(lorenz_c, [-8,8,27], t)
     >>> u_eval = u(t)
     >>> model = SINDy()
     >>> model.fit(x, u_eval, t=t[1]-t[0])
@@ -175,6 +185,13 @@ class SINDy(BaseEstimator):
         multiple_trajectories=False,
         unbias=True,
         quiet=False,
+        ensemble=False,
+        library_ensemble=False,
+        replace=True,
+        n_candidates_to_drop=1,
+        n_subset=None,
+        n_models=None,
+        ensemble_aggregator=None,
     ):
         """
         Fit a SINDy model.
@@ -234,6 +251,48 @@ class SINDy(BaseEstimator):
         quiet: boolean, optional (default False)
             Whether or not to suppress warnings during model fitting.
 
+        ensemble : boolean, optional (default False)
+            This parameter is used to allow for "ensembling", i.e. the
+            generation of many SINDy models (n_models) by choosing a random
+            temporal subset of the input data (n_subset) for each sparse
+            regression. This often improves robustness because averages
+            (bagging) or medians (bragging) of all the models are usually
+            quite high-performing. The user can also generate "distributions"
+            of many models, and calculate how often certain library terms
+            are included in a model.
+
+        library_ensemble : boolean, optional (default False)
+            This parameter is used to allow for "library ensembling",
+            i.e. the generation of many SINDy models (n_models) by choosing
+            a random subset of the candidate library terms to truncate. So,
+            n_models are generated by solving n_models sparse regression
+            problems on these "reduced" libraries. Once again, this often
+            improves robustness because averages (bagging) or medians
+            (bragging) of all the models are usually quite high-performing.
+            The user can also generate "distributions" of many models, and
+            calculate how often certain library terms are included in a model.
+
+        replace : boolean, optional (default True)
+            If ensemble true, whether or not to time sample with replacement.
+
+        n_candidates_to_drop : int, optional (default 1)
+            Number of candidate terms in the feature library to drop during
+            library ensembling.
+
+        n_subset : int, optional (default len(time base))
+            Number of time points to use for ensemble
+
+        n_models : int, optional (default 20)
+            Number of models to generate via ensemble
+
+        ensemble_aggregator : callable, optional (default numpy.median)
+            Method to aggregate model coefficients across different samples.
+            This method argument is only used if ``ensemble`` or ``library_ensemble``
+            is True.
+            The method should take in a list of 2D arrays and return a 2D
+            array of the same shape as the arrays in the list.
+            Example: :code:`lambda x: np.median(x, axis=0)`
+
         Returns
         -------
         self: a fitted :class:`SINDy` instance
@@ -252,6 +311,25 @@ class SINDy(BaseEstimator):
             )
             self.n_control_features_ = u.shape[1]
 
+        if (ensemble or library_ensemble) and n_models is None:
+            n_models = 20
+        if ensemble and n_subset is None:
+            n_subset = x.shape[0]
+        if (
+            ensemble
+            and hasattr(self.feature_library, "weak_form")
+            and self.feature_library.weak_form
+            and self.feature_library.is_uniform
+        ):
+            raise ValueError(
+                "Cannot use a uniform grid and ensembling "
+                "together because ensembling will sub-sample the temporal "
+                "grid (only a problem with weak form PDEs)."
+            )
+        if (n_models is not None) and n_models <= 0:
+            raise ValueError("n_models must be a positive integer")
+        if (n_subset is not None) and n_subset <= 0:
+            raise ValueError("n_subset must be a positive integer")
         if multiple_trajectories:
             x, x_dot = self._process_multiple_trajectories(x, t, x_dot)
         else:
@@ -269,12 +347,17 @@ class SINDy(BaseEstimator):
                 else:
                     x_dot = validate_input(x_dot, t)
 
+        # Set ensemble variables
+        self.ensemble = ensemble
+        self.library_ensemble = library_ensemble
+
         # Append control variables
         if self.n_control_features_ > 0:
             x = concatenate((x, u), axis=1)
 
-        # Drop rows where derivative isn't known
-        x, x_dot = drop_nan_rows(x, x_dot)
+        # Drop rows where derivative isn't known unless using weak PDE form
+        if not hasattr(self.feature_library, "weak_form"):
+            x, x_dot = drop_nan_rows(x, x_dot)
 
         if hasattr(self.optimizer, "unbias"):
             unbias = self.optimizer.unbias
@@ -283,20 +366,112 @@ class SINDy(BaseEstimator):
         steps = [("features", self.feature_library), ("model", optimizer)]
         self.model = Pipeline(steps)
 
+        if hasattr(self.feature_library, "temporal_grid"):
+            tgrid = copy(self.feature_library.temporal_grid)
+        else:
+            tgrid = None
+
         action = "ignore" if quiet else "default"
         with warnings.catch_warnings():
             warnings.filterwarnings(action, category=ConvergenceWarning)
             warnings.filterwarnings(action, category=LinAlgWarning)
             warnings.filterwarnings(action, category=UserWarning)
+            pde_library_flag = False
+            if isinstance(self.feature_library, PDELibrary):
+                if self.feature_library.spatial_grid is not None:
+                    pde_library_flag = True
+            if ensemble and not library_ensemble:
+                self.coef_list = []
+                for i in range(n_models):
+                    x_ensemble, x_dot_ensemble = drop_random_rows(
+                        x,
+                        x_dot,
+                        n_subset,
+                        replace,
+                        tgrid,
+                        self.feature_library,
+                        pde_library_flag,
+                    )
+                    self.model.fit(x_ensemble, x_dot_ensemble)
+                    self.coef_list.append(self.model.steps[-1][1].coef_)
+            elif library_ensemble and not ensemble:
+                self.feature_library.library_ensemble = True
+                (self.feature_library).fit(x)
+                n_output_features = self.feature_library.n_output_features_
+                self.coef_list = []
+                for i in range(n_models):
+                    self.feature_library.ensemble_indices = sort(
+                        numpy.random.choice(
+                            range(n_output_features),
+                            n_candidates_to_drop,
+                            replace=False,
+                        )
+                    )
+                    self.model.fit(x, x_dot)
+                    coef_partial = self.model.steps[-1][1].coef_
+                    for j in range(n_candidates_to_drop):
+                        coef_partial = insert(
+                            coef_partial,
+                            self.feature_library.ensemble_indices[j],
+                            0,
+                            axis=-1,
+                        )
+                    self.coef_list.append(coef_partial)
+            elif ensemble and library_ensemble:
+                self.feature_library.library_ensemble = True
+                (self.feature_library).fit(x)
+                n_output_features = self.feature_library.n_output_features_
+                self.coef_list = []
+                for i in range(n_models):
+                    x_ensemble, x_dot_ensemble = drop_random_rows(
+                        x,
+                        x_dot,
+                        n_subset,
+                        replace,
+                        tgrid,
+                        self.feature_library,
+                        isinstance(self.feature_library, PDELibrary),
+                    )
+                    for j in range(n_models):
+                        self.feature_library.ensemble_indices = sort(
+                            numpy.random.choice(
+                                range(n_output_features),
+                                n_candidates_to_drop,
+                                replace=False,
+                            )
+                        )
+                        self.model.fit(x_ensemble, x_dot_ensemble)
+                        coef_partial = self.model.steps[-1][1].coef_
+                        for k in range(n_candidates_to_drop):
+                            coef_partial = insert(
+                                coef_partial,
+                                self.feature_library.ensemble_indices[k],
+                                0,
+                                axis=-1,
+                            )
+                        self.coef_list.append(coef_partial)
+            else:
+                self.model.fit(x, x_dot)
 
-            self.model.fit(x, x_dot)
+        # Get average coefficients if ensembling was used
+        if ensemble or library_ensemble:
+            if ensemble_aggregator is None:
+                self.model.coef_ = median(self.coef_list, axis=0)
+            else:
+                self.model_coef_ = ensemble_aggregator(self.coef_list)
 
-        self.n_input_features_ = self.model.steps[0][1].n_input_features_
+        # New version of sklearn changes attribute name
+        if float(__version__[:3]) >= 1.0:
+            self.n_features_in_ = self.model.steps[0][1].n_features_in_
+            n_input_features = self.model.steps[0][1].n_features_in_
+        else:
+            self.n_input_features_ = self.model.steps[0][1].n_input_features_
+            n_input_features = self.model.steps[0][1].n_input_features_
         self.n_output_features_ = self.model.steps[0][1].n_output_features_
 
         if self.feature_names is None:
             feature_names = []
-            for i in range(self.n_input_features_ - self.n_control_features_):
+            for i in range(n_input_features - self.n_control_features_):
                 feature_names.append("x" + str(i))
             for i in range(self.n_control_features_):
                 feature_names.append("u" + str(i))
@@ -382,7 +557,9 @@ class SINDy(BaseEstimator):
         else:
             base_feature_names = self.feature_names
         return equations(
-            self.model, input_features=base_feature_names, precision=precision
+            self.model,
+            input_features=base_feature_names,
+            precision=precision,
         )
 
     def print(self, lhs=None, precision=3):
@@ -398,11 +575,21 @@ class SINDy(BaseEstimator):
             Precision to be used when printing out model coefficients.
         """
         eqns = self.equations(precision)
+        if isinstance(self.feature_library, SINDyPILibrary):
+            feature_names = self.get_feature_names()
+        else:
+            feature_names = self.feature_names
         for i, eqn in enumerate(eqns):
             if self.discrete_time:
-                print(self.feature_names[i] + "[k+1] = " + eqn)
+                names = "(" + feature_names[i] + ")"
+                print(names + "[k+1] = " + eqn)
             elif lhs is None:
-                print(self.feature_names[i] + "' = " + eqn)
+                if not isinstance(self.feature_library, SINDyPILibrary):
+                    names = "(" + feature_names[i] + ")"
+                    print(names + "' = " + eqn)
+                else:
+                    names = feature_names[i]
+                    print(names + " = " + eqn)
             else:
                 print(lhs[i] + " = " + eqn)
 
@@ -489,7 +676,8 @@ class SINDy(BaseEstimator):
                 x, t, x_dot, return_array=True
             )
         else:
-            x = validate_input(x, t)
+            if not hasattr(self.feature_library, "weak_form"):
+                x = validate_input(x, t)
             if x_dot is None:
                 if self.discrete_time:
                     x_dot = x[1:]
@@ -505,7 +693,10 @@ class SINDy(BaseEstimator):
             x = concatenate((x, u), axis=1)
 
         # Drop rows where derivative isn't known (usually endpoints)
-        x, x_dot = drop_nan_rows(x, x_dot)
+        if not hasattr(self.feature_library, "weak_form"):
+            x, x_dot = drop_nan_rows(x, x_dot)
+        else:
+            self.feature_library.temporal_grid = t
 
         x_dot_predict = self.model.predict(x)
         return metric(x_dot, x_dot_predict, **metric_kws)
@@ -664,10 +855,10 @@ class SINDy(BaseEstimator):
         x0,
         t,
         u=None,
-        integrator=odeint,
+        integrator="solve_ivp",
         stop_condition=None,
         interpolator=None,
-        integrator_kws={},
+        integrator_kws={"method": "LSODA", "rtol": 1e-12, "atol": 1e-12},
         interpolator_kws={},
     ):
         """
@@ -696,9 +887,10 @@ class SINDy(BaseEstimator):
             u should be a list (with ``len(u) == t``) or array (with
             ``u.shape[0] == 1``) giving the control inputs at each step.
 
-        integrator: callable, optional (default ``odeint``)
+        integrator: string, optional (default ``solve_ivp``)
             Function to use to integrate the system.
-            Default is ``scipy.integrate.odeint``.
+            Default is ``scipy.integrate.solve_ivp``. The only options
+            currently supported are solve_ivp and odeint.
 
         stop_condition: function object, optional
             If model is in discrete time, optional function that gives a
@@ -740,7 +932,11 @@ class SINDy(BaseEstimator):
                 def check_stop_condition(xi):
                     pass
 
-            x = zeros((t, self.n_input_features_ - self.n_control_features_))
+            # New version of sklearn changes attribute name
+            if float(__version__[:3]) >= 1.0:
+                x = zeros((t, self.n_features_in_ - self.n_control_features_))
+            else:
+                x = zeros((t, self.n_input_features_ - self.n_control_features_))
             x[0] = x0
 
             if u is None or self.n_control_features_ == 0:
@@ -773,7 +969,7 @@ class SINDy(BaseEstimator):
                         " were not used when the model was fit"
                     )
 
-                def rhs(x, t):
+                def rhs(t, x):
                     return self.predict(x[newaxis, :])[0]
 
             else:
@@ -796,15 +992,26 @@ class SINDy(BaseEstimator):
 
                 if ndim(u_fun(t[0])) == 1:
 
-                    def rhs(x, t):
+                    def rhs(t, x):
                         return self.predict(x[newaxis, :], u_fun(t).reshape(1, -1))[0]
 
                 else:
 
-                    def rhs(x, t):
+                    def rhs(t, x):
                         return self.predict(x[newaxis, :], u_fun(t))[0]
 
-            return integrator(rhs, x0, t, **integrator_kws)
+            # Need to hard-code below, because odeint and solve_ivp
+            # have different syntax and integration options.
+            if integrator == "solve_ivp":
+                return (
+                    (solve_ivp(rhs, (t[0], t[-1]), x0, t_eval=t, **integrator_kws)).y
+                ).T
+            elif integrator == "odeint":
+                if integrator_kws.get("method") == "LSODA":
+                    integrator_kws = {}
+                return odeint(rhs, x0, t, tfirst=True, **integrator_kws)
+            else:
+                raise ValueError("Integrator not supported, exiting")
 
     @property
     def complexity(self):
