@@ -2,10 +2,12 @@ from itertools import repeat
 from typing import Sequence
 
 import numpy as np
+from numpy.random import choice
 from scipy.optimize import bisect
 from sklearn.base import MultiOutputMixin
 from sklearn.utils.validation import check_array
 
+
 # Define a special object for the default value of t in
 # validate_input. Normally we would set the default
 # value of t to be None, but it is possible for the user
@@ -100,6 +102,100 @@ def drop_nan_rows(x, x_dot):
     return x, x_dot
 
 
+def drop_random_rows(
+    x, x_dot, n_subset, replace, tgrid, feature_library, is_pde_library
+):
+    # Can't choose random n_subset points if data is from a PDE
+    # (and therefore is spatially local).
+    # Need to unfold it and just choose n_subset from the temporal slices
+    if is_pde_library:
+        spatial_grid = feature_library.spatial_grid
+        num_gridx = (spatial_grid).shape[0]
+        if len(np.shape(spatial_grid)) == 1:
+            num_time = np.shape(x)[0] // num_gridx
+            if n_subset > num_time:
+                n_subset = num_time
+            # Weak form needs uniform, ascending grid, so cannot replace
+            if feature_library.weak_form:
+                replace = False
+            rand_inds = np.sort(choice(range(num_time), n_subset, replace=replace))
+            x_shaped = np.reshape(x, (num_gridx, num_time, x.shape[1]))
+            x_shaped = x_shaped[:, rand_inds, :]
+            x_new = np.reshape(x_shaped, (num_gridx * n_subset, x.shape[1]))
+            if not feature_library.weak_form:
+                x_dot_shaped = np.reshape(x_dot, (num_gridx, num_time, x.shape[1]))
+                x_dot_shaped = x_dot_shaped[:, rand_inds, :]
+                x_dot_new = np.reshape(x_dot_shaped, (num_gridx * n_subset, x.shape[1]))
+            elif feature_library.weak_form:
+                x_dot_new = x_dot
+                feature_library.temporal_grid = tgrid[rand_inds]
+        if len(np.shape(spatial_grid)) == 3:
+            num_gridy = (spatial_grid).shape[1]
+            num_time = np.shape(x)[0] // num_gridx // num_gridy
+            if n_subset > num_time:
+                n_subset = num_time
+            # Weak form needs uniform, ascending grid, so cannot replace
+            if feature_library.weak_form:
+                replace = False
+            rand_inds = np.sort(choice(range(num_time), n_subset, replace=replace))
+            x_shaped = np.reshape(x, (num_gridx, num_gridy, num_time, x.shape[1]))
+            x_shaped = x_shaped[:, :, rand_inds, :]
+            x_new = np.reshape(x_shaped, (num_gridx * num_gridy * n_subset, x.shape[1]))
+            if not feature_library.weak_form:
+                x_dot_shaped = np.reshape(
+                    x_dot, (num_gridx, num_gridy, num_time, x.shape[1])
+                )
+                x_dot_shaped = x_dot_shaped[:, :, rand_inds, :]
+                x_dot_new = np.reshape(
+                    x_dot_shaped, (num_gridx * num_gridy * n_subset, x.shape[1])
+                )
+            elif feature_library.weak_form:
+                x_dot_new = x_dot
+                feature_library.temporal_grid = tgrid[rand_inds]
+        if len(np.shape(spatial_grid)) == 4:
+            num_gridy = (spatial_grid).shape[1]
+            num_gridz = (spatial_grid).shape[2]
+            num_time = np.shape(x)[0] // num_gridx // num_gridy // num_gridz
+            if n_subset > num_time:
+                n_subset = num_time
+            # Weak form needs uniform, ascending grid, so cannot replace
+            if feature_library.weak_form:
+                replace = False
+            rand_inds = np.sort(choice(range(num_time), n_subset, replace=replace))
+            x_shaped = np.reshape(
+                x, (num_gridx, num_gridy, num_gridz, num_time, x.shape[1])
+            )
+            x_shaped = x_shaped[:, :, :, rand_inds, :]
+            x_new = np.reshape(
+                x_shaped, (num_gridx * num_gridy * num_gridz * n_subset, x.shape[1])
+            )
+            if not feature_library.weak_form:
+                x_dot_shaped = np.reshape(
+                    x_dot, (num_gridx, num_gridy, num_gridz, num_time, x.shape[1])
+                )
+                x_dot_shaped = x_dot_shaped[:, :, :, rand_inds, :]
+                x_dot_new = np.reshape(
+                    x_dot_shaped,
+                    (num_gridx * num_gridy * num_gridz * n_subset, x.shape[1]),
+                )
+            elif feature_library.weak_form:
+                x_dot_new = x_dot
+                feature_library.temporal_grid = tgrid[rand_inds]
+    else:
+        # Choose random n_subset points to use
+        rand_inds = np.sort(choice(range(np.shape(x)[0]), n_subset, replace=replace))
+        x_new = x[rand_inds, :]
+        if hasattr(feature_library, "weak_form"):
+            if feature_library.weak_form:
+                x_dot_new = x_dot
+                feature_library.temporal_grid = tgrid[rand_inds]
+            else:
+                x_dot_new = x_dot[rand_inds, :]
+        else:
+            x_dot_new = x_dot[rand_inds, :]
+    return x_new, x_dot_new
+
+
 def reorder_constraints(c, n_features, output_order="row"):
     """Reorder constraint matrix."""
     ret = c.copy()
@@ -142,7 +238,17 @@ def prox_l1(x, threshold):
 
 def prox_weighted_l1(x, thresholds):
     """Proximal operator for weighted l1 regularization."""
-    return np.sign(x) * np.maximum(np.abs(x) - thresholds, np.ones(x.shape))
+    return np.sign(x) * np.maximum(np.abs(x) - thresholds, np.zeros(x.shape))
+
+
+def prox_l2(x, threshold):
+    """Proximal operator for ridge regularization."""
+    return 2 * threshold * x
+
+
+def prox_weighted_l2(x, thresholds):
+    """Proximal operator for ridge regularization."""
+    return 2 * thresholds * x
 
 
 # TODO: replace code block with proper math block
@@ -170,14 +276,17 @@ def prox_cad(x, lower_threshold):
 
 
 def get_prox(regularization):
-    if regularization.lower() in ("l0", "weighted_l0"):
-        return prox_l0
-    elif regularization.lower() in ("l1", "weighted_l1"):
-        return prox_l1
-    elif regularization.lower() == "weighted_l1":
-        return prox_weighted_l1
-    elif regularization.lower() == "cad":
-        return prox_cad
+    prox = {
+        "l0": prox_l0,
+        "weighted_l0": prox_weighted_l0,
+        "l1": prox_l1,
+        "weighted_l1": prox_weighted_l1,
+        "l2": prox_l2,
+        "weighted_l2": prox_weighted_l2,
+        "cad": prox_cad,
+    }
+    if regularization.lower() in prox.keys():
+        return prox[regularization.lower()]
     else:
         raise NotImplementedError("{} has not been implemented".format(regularization))
 
@@ -191,6 +300,10 @@ def get_regularization(regularization):
         return lambda x, lam: lam * np.sum(np.abs(x))
     elif regularization.lower() == "weighted_l1":
         return lambda x, lam: np.sum(np.abs(lam @ x))
+    elif regularization.lower() == "l2":
+        return lambda x, lam: lam * np.sum(x ** 2)
+    elif regularization.lower() == "weighted_l2":
+        return lambda x, lam: np.sum(lam @ x ** 2)
     else:
         raise NotImplementedError("{} has not been implemented".format(regularization))
 
