@@ -48,7 +48,7 @@ def test_complexity(n_samples, n_features, n_informative, random_state):
         )
         y = y.reshape(-1, 1)
 
-        opt_kwargs = dict(fit_intercept=True, normalize=False)
+        opt_kwargs = dict(fit_intercept=True)
         optimizers = [
             SR3(thresholder="l0", threshold=0.1, **opt_kwargs),
             SR3(thresholder="l1", threshold=0.1, **opt_kwargs),
@@ -111,4 +111,5 @@ def test_complexity_parameter(
         opt.fit(x, y)
 
     for less_complex, more_complex in zip(optimizers, optimizers[1:]):
+        print(less_complex.complexity, more_complex.complexity)
         assert less_complex.complexity <= more_complex.complexity
